% Crude driver for eigenfrequencies and eigenfunctions of triangles.
% Based upon the line billiard in vergini, and Fourier-Bessel sine basis at
% the corner at the origin. This is not very accurate (3 digits?).
% Around 2% of eigenvalues are missed at k~100, less for higher k.
% Requires verg executable in matlab's path (verg now uses openmp)
% Barnett 9/18/12
%
% This is faster (by factor at least 2), but less accurate (2 digits rather
% than say 4 digits) than gen_tri_mfs.m from 2006. The latter also doesn't
% miss eigenvalues so often. For narrow triangles (a~1, b<<1), it will be
% much faster.
clear

% problem params:
a = 0.7; b = 0.8;  % shape: triangle has corners (0,0), (1,0), and (a,b), a,b>0
k = [100 110];     % k range desired; k=sqrt(eigenvalue). takes 16 s on i7 CPU
% k = [400 405]; or try [6 100] to go from bottom of spectrum

% numerical params
wantEFs = 1;  % 0 if don't want grid of EFs to be computed (slower), 1 if do
kwind = 0.3;   % half k window size: controls how many EVs collected per call
opts.b = 10;   % points per wavelength on boundary
opts.v = 0;    % -2,-1,..,2 amount of matlab text output
opts.bb = [0 max(a,1) 0 b]; % bounding box
opts.dx = 0.01;  % choose to be ~1/k to resolve waves
opts.fast = 0;   % 0 for best EFs; 1 is many times faster but dilated by a
                 % fraction up to kwind/k, so beware (reducing kwind helps).

area = b/2; perim = 1+sqrt(a^2+b^2)+sqrt((1-a)^2+b^2);
bil = sprintf('-l line:%.15g:%.15g:1:0 ',a,b);
% basis params: note, may have to tweak for diff (a,b) choices...
sys = [bil sprintf('-s fb:%g:0:%.15g -u -4 4',1.5+3/sqrt(k(2)),atan(b/a)/pi)];

tic % Compute eigenfreqs (ks) & bdry normal-deriv data on line (a,b) to (1,0):
if wantEFs
  [ks, prop, err, bdry, g] = gen_levels(k(1), k(2), kwind, kwind, sys, opts);
else
  [ks, prop, err, bdry] = gen_levels(k(1), k(2), kwind, kwind, sys, opts);
end
toc, ne = numel(ks)

if 1 % find and remove duplicates ---------------
norms = sqrt(dot(bdry.ngr,bdry.ngr,2));
diago = sum(bdry.ngr(1:end-1,:).*bdry.ngr(2:end,:), 2);
coso = diago ./ (norms(1:end-1).*norms(2:end));    % normalize to get cos angs
j_dups = find(abs(coso)>0.99)                 % adjust cutoff dep on acc
ind = ones(1,ne); ind(j_dups) = 0;
j_keep = find(ind); ne = numel(j_keep);
ks = ks(j_keep); bdry.ngr = bdry.ngr(j_keep,:); bdry.per = bdry.per(j_keep,:);
err.ten = err.ten(j_keep);err.nrm = err.nrm(j_keep);err.kos = err.kos(j_keep);
if wantEFs, g.sta = g.sta(j_keep,:,:); end
end % ------

% check missing eigenvalues (missing whenever graph goes up by 1):
figure; weyl(0, ks, perim, area); title('visual check missing eigvals');
% don't use prop.area since wrong
figure; semilogy(ks, err.ten, '+'); title('tension values (error^2) vs k');

if 1 % zero the EF values outside the triangle if you want...
  [xx yy] = meshgrid(g.xs,g.ys);
  mask = -b*xx+a*yy<=0 & -b*(xx-1)+(a-1)*yy>=0; % test upper 2 edges
  g.sta = g.sta .* repmat(reshape(mask', [1 g.nx g.ny]), [ne 1 1]);
end
  
if wantEFs
  figure; i=1; imagesc(g.xs,g.ys,squeeze(g.sta(i,:,:))'); % plot i'th EF
  axis xy equal tight; caxis([-2 2]);
  
  figure; imagesc(g.xs,g.ys,squeeze(sum(g.sta.^2,1))'); axis xy equal tight;
  colorbar;  % basic stats, eg show sum of squares of all EFs found
  
  %figure; n=ceil(sqrt(2*ne)); for i=1:ne,  % plot all EFs in one fig
  %  subplot(ceil(ne/n), n, i); imagesc(g.xs,g.ys,squeeze(g.sta(i,:,:))');
  %axis xy equal tight off; caxis([-2 2]); end
  
  save_sta(g.sta, ks, 't'); % save viewer-readable t.sta_bin file...
end

% save as matlab...
%save tria.7b.8k100.mat

% random debugging stuff:
%[sta, k, ne, nx, ny] = load_sta('tria.7b.8k200');
%imagesc(g.xs,g.ys,squeeze(g.sta(i,:,:)-sta(i,:,:))');
%axis xy equal tight; caxis([-2 2]);
